﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Mail;
using System.Security.Cryptography;
using System.Text;
using System.Threading.Tasks;
using Org.BouncyCastle.Asn1.Pkcs;
using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Generators;
using Org.BouncyCastle.Crypto.Prng;
using Org.BouncyCastle.OpenSsl;
using Org.BouncyCastle.Security;

namespace Cryptography.Classes
{
	public class RsaKey
	{
		#region Public Properties
		public string privateKey { get; set; }
		public string publicKey { get; set; }
		#endregion

		#region Public Methods
		public string encrypt (
			string data,
			bool usePublicKey = false
		)
		{
			var rsa = this.asRSACryptoServiceProvider(usePublicKey: usePublicKey);

			byte[] encryptedData = rsa.Encrypt(System.Text.Encoding.UTF8.GetBytes(data), true);
			var str = Convert.ToBase64String(encryptedData);

			return (str);
		}

		public string decrypt (
			string data,
			bool usePublicKey = false
		)
		{
			var rsa = this.asRSACryptoServiceProvider(usePublicKey: usePublicKey);

			byte[] decryptedData = rsa.Decrypt(Convert.FromBase64String(data), true);
			var str = System.Text.Encoding.UTF8.GetString(decryptedData);

			return (str);
		}

		public string hash (
			string data,
			bool usePublicKey = false
		)
		{
			var rsa = this.asRSACryptoServiceProvider(usePublicKey: usePublicKey);

			byte[] encryptedData = rsa.SignData(System.Text.Encoding.UTF8.GetBytes(data), new SHA512CryptoServiceProvider());
			var str = Convert.ToBase64String(encryptedData);

			return (str);
		}

		public void verifyHash (
			string data,
			string hash,
			bool usePublicKey = false
		)
		{
			var rsa = this.asRSACryptoServiceProvider(usePublicKey: usePublicKey);

			var isValid = rsa.VerifyData(System.Text.Encoding.UTF8.GetBytes(data), new SHA512CryptoServiceProvider(), Convert.FromBase64String(hash));
			if (!isValid)
			{
				throw new CryptographicException("Invalid hash.");
			}
		}

		public RSACryptoServiceProvider asRSACryptoServiceProvider (
			bool usePublicKey = false
		)
		{
			var keyStr = usePublicKey ? this.publicKey : this.privateKey;

			var pemReader = new PemReader(new StringReader(keyStr));

			if (usePublicKey)
			{
				var asymmetricKey = (AsymmetricKeyParameter) pemReader.ReadObject();
				pemReader.Reader.Close();

				var rsaKey = (Org.BouncyCastle.Crypto.Parameters.RsaKeyParameters) asymmetricKey;
				var rsaKeyInfo = Org.BouncyCastle.Security.DotNetUtilities.ToRSAParameters(rsaKey);

				var rsa = new RSACryptoServiceProvider();
				rsa.ImportParameters(rsaKeyInfo);

				return (rsa);
			}
			else
			{
				var pemObject = pemReader.ReadObject();
				Org.BouncyCastle.Crypto.Parameters.RsaPrivateCrtKeyParameters rsaKey;
				if (pemObject is Org.BouncyCastle.Crypto.Parameters.RsaPrivateCrtKeyParameters)
				{
					rsaKey = (Org.BouncyCastle.Crypto.Parameters.RsaPrivateCrtKeyParameters) pemObject;
				}
				else
				{
					var asymmetricKey = (AsymmetricCipherKeyPair) pemObject;
					rsaKey = (Org.BouncyCastle.Crypto.Parameters.RsaPrivateCrtKeyParameters) asymmetricKey.Private;
				}
				pemReader.Reader.Close();

				var rsaKeyInfo = Org.BouncyCastle.Security.DotNetUtilities.ToRSAParameters(rsaKey);

				var rsa = new RSACryptoServiceProvider();
				rsa.ImportParameters(rsaKeyInfo);

				return (rsa);

			}
		}

		public static RsaKey generateKeyPair (
			string signatureAlgorithm = null,
			int rsaKeyLength = 2048
		)
		{
			if (string.IsNullOrEmpty(signatureAlgorithm))
			{
				signatureAlgorithm = PkcsObjectIdentifiers.Sha512WithRsaEncryption.Id;
			}

			var rsaKeyPairGenerator = new RsaKeyPairGenerator();
			rsaKeyPairGenerator.Init(new KeyGenerationParameters(new SecureRandom(new CryptoApiRandomGenerator()), rsaKeyLength));

			var pair = rsaKeyPairGenerator.GenerateKeyPair();
			var configuration = new RsaKey();

			var privateKeyStrBuilder = new StringBuilder();
			var privateKeyPemWriter = new PemWriter(new StringWriter(privateKeyStrBuilder));
			privateKeyPemWriter.WriteObject(pair.Private);
			privateKeyPemWriter.Writer.Flush();
			privateKeyPemWriter.Writer.Close();

			configuration.privateKey = privateKeyStrBuilder.ToString();


			var publicKeyStrBuilder = new StringBuilder();
			var publicKeyPemWriter = new PemWriter(new StringWriter(publicKeyStrBuilder));
			publicKeyPemWriter.WriteObject(pair.Public);
			publicKeyPemWriter.Writer.Flush();
			publicKeyPemWriter.Writer.Close();

			configuration.publicKey = publicKeyStrBuilder.ToString();

			return (configuration);
		}
		#endregion
	}
}
